/* <X>SCC-Driver for PC-FlexNet               (C) 2000 H. de Groot PE1DNN */

#include "stdio.h"
#include "dos.h"
#include "ctype.h"
#include "stdlib.h"
#include "string.h"
#include "FLEXDRV.H"

static char version[] = "0.23";

static char helptext[] = "\n\n"
"Driver for OSCC, PSCC, FSCC, DRSI, DSCC and USCC cards\n\n"
"Usage:\n\n"
"XSCC -t:<type> -p:<baseaddr> -i:<irq> -c:<channels> -x:<xtal-fr> -d:<time>\n\n"
"-t:<type>    = OSCC, PSCC, FSCC, DRSI, DSCC or USCC default OSCC (PA0HZP)\n"
"-p:<baseaddr>= I/O address, default 0x150 (OSCC/PSCC/FSCC) or 0x300 (others)\n"
"-i:<irq>     = IRQ of SCC card, default 3 (OSCC/PSCC/FSCC) or 7 (others)\n"
"-c:<channels>= Number of channels you want to use on the card, shall be\n"
"               equal or less then the number of channels supported by the\n"
"               card. Default is maximum number of channels of the card.\n"
"-x:<xtal-fr> = Xtal frequency on the SCC card, default 4.9152\n"
"               enter freq in MHZ [example -x:4.9152]\n"
"-d:<time>    = PTT TX-Tail delay to be used on all full duplex ports.\n"
"               enter time in seconds (0 to 60) [example -d:30]\n";

/* Defines and macro's */
#define MAX_SCCCHANS     4

#define MAX_INTLOOPS  1000

#define PCLK        49152L

#define RX_BUFCOUNT     32
#define TX_BUFCOUNT     16

#define TXEOF       0x8000

#define LOBYTE(val)         ((val) & 0xFF)
#define HIBYTE(val)         ((val) >> 8)

/*** Z8530 ******************************************************************/

/* Write Registers */

#define RES_EX_INT  0x10        /* WR0 */
#define RES_TX_IP   0x28
#define RES_ERR     0x30
#define RES_TX_CRC  0x80
#define RES_EOM     0xC0

#define EX_IE       0x01        /* WR1 */
#define TX_IE       0x02
#define RX_IE       0x10

#define RX_DIS      0xC8        /* WR3 */
#define RX_EN       0xC9

#define SDLC        0x20        /* WR4 */

#define TX_DIS      0x61        /* WR5 */
#define TX_EN       0x69
#define RTS         0x02

#define SDLC_FLAG   0x7E        /* WR7 */

#define MINT_EN     0x0A        /* WR9 */
#define HW_RES      0xC0

#define NRZ         0x80        /* WR10 */
#define NRZI        0xA0
#define ABUNDER     0x04

#define TRxC_RTxC   0x20        /* WR11 */
#define DPLL_RTxC   0x64
#define DPLL_BRG    0x74
#define TRxOUT_BRG  0x02
#define TRxOUT_DPLL 0x03

#define DPLL_EN     0x23        /* WR14 */
#define DPLL_DIS    0x60
#define DPLL_SRCBRG 0x80
#define DPLL_NRZI   0xE0
#define BRG_EN      0x63

#define ZC_IE       0x02        /* WR15 */
#define DCD_IE      0x08
#define SYNC_IE     0x10
#define CTS_IE      0x20

/* Read Registers */

#define DCD         0x08        /* RR0 */
#define HUNT        0x10
#define CTS         0x20
#define TX_EOM      0x40
#define ABORT       0x80

#define RESIDUE     0x0E        /* RR1 */
#define RES8        0x06
#define RX_OVR      0x20
#define CRC_ERR     0x40
#define SDLC_EOF    0x80

#define SCCDELAY            inportb (0xE4)      /* non-existing port (2s) */

#define WR0(val)            outportb ((SCCDELAY, scc->ctrl), val)
#define WR8(val)            WR (8, val)
#define WR(reg, val)        outportb ((SCCDELAY, outportb (scc->ctrl, reg), \
                                   SCCDELAY, scc->ctrl), val)

#define RR0                 inportb ((SCCDELAY, scc->ctrl))
#define RR8                 inportb ((SCCDELAY, scc->data))
#define RR(reg)             inportb ((SCCDELAY, outportb (scc->ctrl, reg), \
                                      SCCDELAY, scc->ctrl))

/* Ports */
#define PIC0        0x0020        /* Master PIC */
#define PIC1        0x00A0        /* Slave PIC */

#define EOI         0x0020        /* end of interrupt */
#define SETPRIO     0x00C0        /* set IRQ priority */

#define IRQ0        0x0008        /* int vector of IRQ0 */
#define IRQ8        0x0070        /* int vector of IRQ8 */

/* IRQ control */
#define irqvec(irq)         (irq + (irq < 8 ? IRQ0 : IRQ8-8))

#define maskreg(irq)        (irq < 8 ? PIC0+1 : PIC1+1)

#define irqmask(irq)        (1 << (irq & 7))

#define eoi(irq)            if (irq >= 8) \
                                outportb (PIC1, EOI); \
                            outportb (PIC0, EOI)

#define idisable()          { _asm pushf; \
                              _asm cli;  }
#define irestore()          { _asm popf; }

/* Local enums, types and structs */

enum {OSCC, PSCC, FSCC, DRSI, DSCC, USCC};                  /* SCC Types    */

enum {DISABLE, AFSK, G3RUH, DF9IC, PA0HZP};                 /* Modem Types  */

enum {IDLE, DELAY, ACTIVE, FLUSH, TAIL, CALIBRATE};         /* TX States    */

enum {FALSE, TRUE};                                         /* For booleans */

typedef byte bool;

/* Buffer pool structure */
typedef struct pool_s {
   struct pool_s *next;         /* Pointer to next buffer in pool or list   */
   L1FRAME        l1frame;      /* L1 frame buffer of FlexNet               */
} pool_t;

/* SCC port administration structure */
typedef struct scc_s {
    bool active;            /* Operational status of port                   */

    byte chan;              /* Channel number                               */
    u16  data, ctrl;        /* Address of SCC Data and Ctrl registers       */
    u16  baud;              /* Baud-rate/100                                */
    byte mode;              /* Mode of the port                             */
    byte modem;             /* Modem connected to the port                  */
    bool dcd;               /* DCD status of port                           */
    byte tx_state;          /* State of the transmit state machine          */
    i16  tx_timer;          /* General timer used for the transmitter       */
    byte cal_timer;         /* Timer used for calibration                   */
    byte rr0;               /* Copy of last read RR0                        */
    bool rx_err;            /* Flag to skip the frame beiing received       */

    /* RX buffers */
    pool_t *rx_pending;     /* RX data waiting to be read by the kernel     */
    pool_t *rx_pending_last;/* end of the RX data chain, used to append     */
    pool_t *rx_to_flexnet;  /* Most recent buffer handed over to the kernel */
    pool_t *rx_receive;     /* Buffer in use by the reception in progress   */
    i16     rx_len;         /* Length of buffer being received              */
    char   *rx_ptr;         /* Pointer in L1FRAME buffer, where to put next */

    /* TX buffers */
    pool_t *tx_pending;     /* TX data waiting to be transmitted by XSCC    */
    pool_t *tx_pending_last;/* end of the TX data chain, used to append     */
    pool_t *tx_transmit;    /* TX buffer of transmission in progress        */
    i16     tx_len;         /* Nr of bytes to transmit from buffer          */
    char   *tx_ptr;         /* Pointer in L1FRAME buffer, char to TX next   */
} scc_t;

/* Structure of SCC board information and defaults */
typedef struct sccboard_s {
    u16  base;                  /* default base I/O address of the SCC card */
    byte irq;                   /* default irq address of the SCC card      */
    byte data[MAX_SCCCHANS];    /* offset of data-regs w.r.t. base I/O addr */
    u16  ctrl;                  /* offset of ctrl-reg w.r.t. data register  */
    u16  chans;                 /* number of channels supported by the card */
} sccboard_t;

typedef void (interrupt far *procfp) (void);

/* Local variables */
pool_t rx_bufpool[RX_BUFCOUNT]; /* Pool of receive buffers                  */
pool_t *next_free_rx;           /* Next free buffer in the RX buffer pool   */
byte rx_channel;                /* Channel used in last RX query            */
pool_t tx_bufpool[TX_BUFCOUNT]; /* Pool of transmit buffers                 */
pool_t *next_free_tx;           /* Next free buffer in the TX buffer pool   */
pool_t *tx_to_flexnet;          /* Most recent TX buffer given to kernel    */
byte tx_channel;                /* Channel for tx_to_flexnet buffer         */

u16  tail_time;                 /* TX-Tail time for full-duplex             */

void int_tx (scc_t *);          /* SCC transmit interrupt                   */
void int_ex (scc_t *);          /* SCC extern status change interupt        */
void int_rx (scc_t *);          /* SCC receive interrtupt                   */
void int_sp (scc_t *);          /* SCC special condition interrupt          */

/* SCC board defaults */
sccboard_t sccboard[] = {
/* base,irq,{data reg-offsets}   ctrl-offset, chans */
{ 0x150, 3, {3, 1, 7, 5},             -1,       4},         /* OSCC  */
{ 0x150, 3, {3, 1},                   -1,       2},         /* PSCC  */
{ 0x150, 3, {3, 1, 7, 5},             -1,       4},         /* FSCC  */
{ 0x300, 7, {3, 1},                   -1,       2},         /* DRSI  */
{ 0x300, 7, {1, 0, 3, 2},             16,       4},         /* DSCC  */
{ 0x300, 7, {0, 1, 2, 3},              4,       4}};        /* USCC  */

scc_t scctab[MAX_SCCCHANS];
scc_t *sccend     = scctab;                         /* set end to start     */
byte scctype      = OSCC;                           /* default SCC type     */
byte sccchans     = MAX_SCCCHANS + 1;               /* illegal value        */
byte sccirq       = 0;                              /* illegal IRQ          */
byte max_sccchans;                                  /* max. nr of channels  */
u32  pclk         = PCLK;                           /* SCC clock (Xtal)     */

bool interrupts_initialised = FALSE;                /* ints not initialised */
bool driver_initialised     = FALSE;                /* driver not initialised */

procfp old_sccint, old_int_1C;   /* interrupt handlers for SCC and PC timer */

void (*handler[]) (scc_t *) = {int_tx, int_ex, int_rx, int_sp};

/* Channel diagnosis: Allocate one array element for every channel. Needs
   only be as large as the number of defined channels */
static L1_STATISTICS statistics[MAX_SCCCHANS];

/* ------------------------------------------------------------------------- */
/* init_buffers - intitialise the RX and TX buffer pool */
static void init_buffers(void)
{
    int i;

    /*
     * On both buffer pools prepare the following chain:
     *
     *            +-------+
     *      +-----|next   |<----- next_free_rx or next_free_tx
     *      |     |l2frame|
     *      |     +-------+
     *      +---->|   next|-----+
     *            |l2frame|     |
     *            +-------+     |
     *      +-----|next   |<----+
     *      |     |l2frame|
     *      |     +-------+
     *      +---->|   next|-----> NULL
     *            |l2frame|
     *            +-------+
     *
     * When a buffer is allocated then the pointer for the buffer is set to
     * next_free_?x, and next_free_?x will become next_free_?x->next. (where
     * ? is 'r' or 't' depending on which list is used).
     *
     * If next_free_?x is NULL then all buffers are used up.
     *
     * When a buffer is freed, the next pointer of the buffer is set to
     * next_free_xx and next_free_xx will then become a pointer to the freed
     * buffer, so buffers are added at the front (because that's the most
     * simple).
     *
     * Also the rx_pending and tx_pending list will have this structure.
     * Those list will have an extra the pointer to the end which will
     * be rx_pending_last and tx_pending_last. It is used to append new
     * buffers, since for this list the order is important (shall work as
     * FIFO, also called queue).
     *
     * Note that operations on the linked lists need to be atomic, if it gets
     * interrupted results are disastrous.
     */

    /* Initialise RX pool */
    next_free_rx = &rx_bufpool[0];
    for(i = 0; i < (RX_BUFCOUNT - 2); i++)
    {
        /* let next pointer point to the next buffer */
        rx_bufpool[i].next = &(rx_bufpool[i + 1]);
    }
    /* the last buffer will point to NULL */
    rx_bufpool[RX_BUFCOUNT-1].next = NULL;

    /* Initialise TX pool */
    next_free_tx = &tx_bufpool[0];
    for(i = 0; i < (TX_BUFCOUNT - 2); i++)
    {
        /* let next pointer point to the next buffer */
        tx_bufpool[i].next = &(tx_bufpool[i + 1]);
    }
    /* the last buffer will point to NULL */
    tx_bufpool[TX_BUFCOUNT - 1].next = NULL;

    tx_to_flexnet = NULL;
    rx_channel    = 0;
}
/* ------------------------------------------------------------------------- */
/* rx_free - free a RX buffer */
/*           shall only be called with disabled interrupts */
static void rx_free(pool_t *ptr)
{
    /* Free receive buffer */
    ptr->next    = next_free_rx;
    next_free_rx = ptr;
}
/* ------------------------------------------------------------------------- */
/* tx_free - free a TX buffer */
/*           shall only be called with disabled interrupts */
static void tx_free(pool_t *ptr)
{
    /* Free transmit buffer */
    ptr->next    = next_free_tx;
    next_free_tx = ptr;
}
/* ------------------------------------------------------------------------- */
/* txget - get character from tx_buffer */
/*         shall only be called with disabled interrupts */
static u16 txget(scc_t *scc)
{
    u16     character;
    pool_t *ptr;

    if(scc->tx_transmit == NULL)
    {
        if(scc->tx_pending == NULL)
        {
            /* no more data pending, return EOF */
            return TXEOF;
        }
        /* get the next buffer */
        ptr              = scc->tx_pending;
        scc->tx_pending  = ptr->next;
        scc->tx_transmit = ptr;
        scc->tx_ptr      = (ptr->l1frame).frame;
        scc->tx_len      = (ptr->l1frame).len;
    }

    if(scc->tx_len > 0)
    {
        scc->tx_len--;
        character = ((u16) (*(scc->tx_ptr))) & 0x00ff;
        scc->tx_ptr++;
        return character;
    }
    else
    {
        /* End of the buffer, free the buffer and return EOF */
        tx_free(scc->tx_transmit);

        /* This will force a fetch of a new buffer on the next call */
        scc->tx_transmit = NULL;
        return TXEOF;
    }
}
/* ------------------------------------------------------------------------- */
/* rxdiscard - discard what is received so far */
/*             shall only be called with disabled interrupts */
static void rxdiscard(scc_t *scc)
{
    /* Throwing away is easy! */
    if(scc->rx_receive != NULL)
    {
        /* set pointer to start and length to 0 */
        scc->rx_ptr = ((scc->rx_receive)->l1frame).frame;
        scc->rx_len = 0;
    }
}
/* ------------------------------------------------------------------------- */
/* rxeof - rx_buffer filled, ready for higher layers */
/*         shall only be called with disabled interrupts */
static void rxeof(scc_t *scc)
{
    L1FRAME far    *p;

    if(scc->rx_receive == NULL)
    {
        /* This is strange! Should point to a buffer */
        return;
    }

    /* Here is the buffer with the now ready "payload"  */
    /* finish it and then append to the rx_pending list */
    p          = &((scc->rx_receive)->l1frame);
    p->len     = scc->rx_len - 1;       /* length of the frame      */
    p->kanal   = scc->chan;             /* received on this channel */
    p->txdelay = 0;                     /* not supported            */

    /* the next pointer shall point NULL, this will be the end of the chain */
    (scc->rx_receive)->next = NULL;

    /* add to the receive list */
    if(scc->rx_pending == NULL)
    {
        /* first buffer in rx_pending list */
        scc->rx_pending = scc->rx_receive;
    }
    else
    {
        /* add buffer to rx_pending list */
        (scc->rx_pending_last)->next = scc->rx_receive;
    }
    /* this is the last frame we added */
    scc->rx_pending_last = scc->rx_receive;

    /* buffer is now in the pending list clear this reference to the buffer */
    scc->rx_receive = NULL;
    scc->rx_len     = 0;

    /* received a frame, note it down */
    statistics[scc->chan].rx_frames++;
}
/* ------------------------------------------------------------------------- */
/* card_type - return verbose name of card-type */
static char *card_type(void)
{
    switch(scctype) {
    case OSCC:  return "Opto SCC (PA0HZP/D.D.S.) ";
    case PSCC:  return "PET SCC (PE1PET) ";
    case FSCC:  return "French SCC (F1FRV/Aeitech) ";
    case DRSI:  return "DRSI SCC ";
    case DSCC:  return "BayCom Digi SCC ";
    case USCC:  return "BayCom USCC ";
    }
    return "Unknown ";
}
/* ------------------------------------------------------------------------- */
/* show_config - dump configuration on the console */
static void show_config(void)
{
    u16   channel;
    scc_t *scc;

    pstr("\nCard type: ");
    pstr(card_type());
    pnum(l1_get_ch_cnt());
    if(l1_get_ch_cnt() == 1)
        pstr(" channel on irq ");
    else
        pstr(" channels on irq ");
    pnum(sccirq);
    pstr(" SCC Xtal ");
    pnum(pclk / 10000);
    pch('.');
    pnum((pclk / 1000) % 10);
    pnum((pclk / 100) % 10);
    pnum((pclk / 10) % 10);
    pnum(pclk % 10);
    pstr(" MHz\n");
    pstr("SCC port settings per SCC chip:\n");
    for (scc = scctab; scc < sccend; scc++)
    {
        pstr((char *)l1_ident(scc->chan));
        pstr(":");
        pnum(scc->chan);
        pstr(": ");
        if(scc->active == TRUE)
        {
            pstr("Data      = 0x");
            phex(scc->data); pstr(",  ");
            pstr("Ctrl    = 0x");
            phex(scc->ctrl); pch('\n');
        }
        else
        {
            pstr("*OFF*\n");
        }
    }
    if(tail_time != 0)
    {
        pstr("Note: a TX-tail delay of ");
        pnum(tail_time / 100);
        pstr(" second");
        if(tail_time != 100)
            pch('s');
        pstr(" will be used on all Full-Duplex ports\n");
    }
}
/* ------------------------------------------------------------------------- */
/* irq_init, init interrupt vector */
static procfp irq_init (u16 irq, procfp handler)
{
    procfp old_int;
    u16  pic, vec;
    byte mask;

    /* map interupt vector to the new handler */
    vec     = irqvec (irq);
    old_int = getvect (vec);
    setvect (vec, handler);

    /* unmask the interrupt in the master or slave 8259 PIC */
    pic     = maskreg (irq);
    mask    = inportb (pic);
    outportb (pic, mask & ~irqmask (irq));

    return old_int;
}
/* ------------------------------------------------------------------------- */
/* irq_reset, reset interrupt vector */
static void irq_reset (byte irq, procfp handler)
{
    u16  pic, vec;
    byte mask;

    /* mask the interrupt in the master or slave 8259 PIC */
    pic     = maskreg (irq);
    mask    = inportb (pic);
    outportb (pic, mask | irqmask (irq));

    /* restore interupt vector of to handler */
    vec     = irqvec (irq);
    setvect (vec, handler);
}
/* ------------------------------------------------------------------------- */
/* scc_clk - program clock speed */
/*           shall only be called with disabled interrupts */
static void scc_clk (scc_t *scc, bool tx_on)
{
    u16 tconst;

    /*
     * Except for AFSK only called when (re)initializing
     * in which case tx_on is always FALSE
     *
     * Since the timing changes between receive and transmit AFSK
     * modem cannot do full duplex. Duplex just means 'disregard DCD' in
     * that case.
     */

    if (scc->modem == DF9IC)
    {
        WR (11, TRxC_RTxC);
        WR (14, DPLL_DIS);
    }
    else
    {
        tconst = pclk / 64L / scc->baud;
        if (tx_on)
            tconst *= 32;
        tconst -= 2;

        WR (12, LOBYTE (tconst));
        WR (13, HIBYTE (tconst));

        WR (11, tx_on ? DPLL_BRG | TRxOUT_BRG
                      : (scc->modem == PA0HZP ? DPLL_RTxC | TRxOUT_BRG
                                              : DPLL_RTxC | TRxOUT_DPLL));

        WR (14, DPLL_SRCBRG);
        WR (14, DPLL_NRZI);
        WR (14, DPLL_EN);
    }
}
/* ------------------------------------------------------------------------- */
/* scc_tx - switch transmitter on/off */
/*          shall only be called with disabled interrupts */
static void scc_tx (scc_t *scc, bool tx_on)
{
    if (scc->modem == AFSK)
    {
        WR (3, RX_DIS);
        WR (5, TX_DIS);

        scc_clk (scc, tx_on);

        if (tx_on)
            WR (5, TX_EN | RTS);
        else
            WR (3, RX_EN);
    }
    else
        WR (5, tx_on ? TX_EN | RTS : TX_EN);
}
/* ------------------------------------------------------------------------- */
/* reset Z8530 SCC */
static void scc_hw_reset (void)
{
    scc_t *scc;
    u16 delay;

    for (scc = scctab; scc < &scctab[max_sccchans]; scc += 2)
    {
        RR0;
        WR (9, HW_RES);
        for (delay = 100; delay--;);
    }
}
/* ------------------------------------------------------------------------- */
/* imergency_reset - something completly wrong, reset the SCC card and driver*/
/*                   shall only be called with disabled interrupts */
static void imergency_reset(void)
{
    byte       channel;
    scc_t      *scc;

    scc_hw_reset ();

    for (scc = scctab; scc < &scctab[max_sccchans]; scc++)
    {
        scc->rx_pending      = NULL;
        scc->rx_pending_last = NULL;
        scc->rx_receive      = NULL;
        scc->rx_to_flexnet   = NULL;
        scc->rx_len          = 0;
        scc->rx_ptr          = NULL;

        scc->tx_pending      = NULL;
        scc->tx_pending_last = NULL;
        scc->tx_transmit     = NULL;
        scc->tx_len          = 0;
        scc->tx_ptr          = NULL;
    }

    /* reinitialise the buffers */
    init_buffers();

    /* driver needs reinitialisation */
    driver_initialised = FALSE;
}
/* ------------------------------------------------------------------------- */
/* scc_int, interrupt entry of the SCC card interrupt */
static void interrupt far scc_int (void)
{
    scc_t *scc;
    byte   rr2;
    u16    watchdog;

    /*
     * If after MAX_INTLOOPS we are still looping take drastic measures
     * to reset the whole card.
     */
    watchdog=MAX_INTLOOPS;

    /*
     * Handle the SCC with disabled interrupts, the Z8530 just does not
     * support concurrent access by design. Also our linked-list buffer
     * manipulation is not re-entrant, so shutting off the interrupts also
     * makes buffer-handling painless within the ISR.
     */
    disable (); /* on iret it will be restored, i.e. at the end of scc_int */
    eoi (sccirq);

    if(driver_initialised == FALSE) return; /* must be left-over interrupt */

    for (scc = scctab; scc < sccend;)
    {
        if (RR (3)) /* Interrupt Pending read on A channel of SCC */
        {
            /* there are interrupts pending for this SCC */
            scc++;         /* advance to B channel */
            rr2 = RR (2);  /* read modified interrupt vector */
            handler[rr2 >> 1 & 3] (&scctab[rr2 >> 3 ^ 1]);
            scc = scctab;
        }
        else
        {
            /*
             * SCC has no interrupts pending (anymore), advance to next
             * SCC (two ports per z8530 SCC chip)
             */
            scc += 2;
        }
        if(watchdog-- == 0)
        {
            /* Serious problem! SCC(s) keeps reporting interrupts */
            /* quit to avoid a complete system hangup.            */

            /*
             * We use the io_error counter to count these events
             * on all involved channels
             */
            for (scc = scctab; scc < sccend; scc++)
                statistics[scc->chan].io_error++;

            /* we reset the whole SCC card and at the first call  */
            /* to l1_init_kanal we initialise from scratch.       */
            imergency_reset();

            return;
        }
    }
}
/* ------------------------------------------------------------------------- */
/* int_tx, handle transmit interrupt of the SCC card */
static void int_tx (scc_t *scc)
{
    u16 val;

    switch (scc->tx_state)
    {
    case ACTIVE:
        if (RR0 & TX_EOM)
        {
            statistics[scc->chan].tx_error++;
            if(scc->tx_transmit != NULL)
            {
                /* set pointer to start and length to lenght of frame */
                scc->tx_ptr = ((scc->tx_transmit)->l1frame).frame;
                scc->tx_len = ((scc->tx_transmit)->l1frame).len;
            }
            scc->tx_state = FLUSH;
        }
        else
            if ((val = txget (scc)) & TXEOF)
            {
                WR (10, (scc->mode & MODE_z) ? NRZ : NRZI);
                WR0 (RES_TX_IP);
                scc->tx_state = FLUSH;
                statistics[scc->chan].tx_frames++;
            }
            else
                WR8 (val);
        break;

    case FLUSH:
        WR (10, (scc->mode & MODE_z) ? NRZ | ABUNDER : NRZI | ABUNDER);

        if ((val = txget (scc)) < TXEOF)
        {
            WR0 (RES_TX_CRC);
            WR8 (val);
            WR0 (RES_EOM);
            scc->tx_state = ACTIVE;
        }
        else
        {
            WR0 (RES_TX_IP);
            if (((scc->mode & MODE_d) == 0) || (tail_time == 0))
            {
                /*
                 * 9600 baud or higher does not need a long tail
                 * Wait for one tick so PTT is switch off in the
                 * part of the code that handles the tx_timer.
                 */
                if( scc->baud >= 96 )
                    scc->tx_timer = 1;
                else
                    scc->tx_timer = 7;
            }
            else
            {
                /*
                 * keep PTT keyed for tail_time seconds. If no new data
                 * is transmitted within this minute the PTT will drop
                 */
                scc->tx_timer = tail_time;
            }
            scc->tx_state = TAIL;
        }
        break;
    default:
        /*
         * this can happen if the driver was reset during transmission,
         * just clear the interrupt, no characters shall follow!
         */
        WR0 (RES_TX_IP);
        break;
    }
}
/* ------------------------------------------------------------------------- */
/* int_ex, handle extern status change interrupt of the SCC card */
static void int_ex (scc_t *scc)
{
    byte rr0, delta;

    delta = (rr0 = RR0) ^ scc->rr0;

    if ((scc->mode & MODE_off) == 0)
    {
        if (scc->rx_len && rr0 & (ABORT | HUNT))
        {
            rxdiscard (scc);
            scc->rx_err = FALSE;
        }

        if(scc->mode & MODE_c)
        {
            /* software DCD */
            if (delta & HUNT)
                scc->dcd = !(rr0 & HUNT);
        }
        else
        {
            /* hardware DCD */
            if (delta & DCD)
                scc->dcd = rr0 & DCD;
        }
    }

    scc->rr0 = rr0;
    WR0 (RES_EX_INT);
}
/* ------------------------------------------------------------------------- */
/* int_rx, handle receive interrupt of the SCC card */
static void int_rx (scc_t *scc)
{
    byte    rr8;
    pool_t *ptr;

    rr8 = RR8;

    if (scc->rx_err == TRUE) return;

    if (scc->rx_receive == NULL)
    {
        /* need to allocate a new buffer */
        if (next_free_rx == NULL)
        {
            /* Ai, out of buffers! We have to skip this frame :-( */
            statistics[scc->chan].rx_bufferoverflow++;
            scc->rx_err = TRUE;
            return;
        }
        /* Allocate receive buffer */
        ptr          = next_free_rx;
        next_free_rx = next_free_rx->next;

        /* starting a new reception */
        scc->rx_receive = ptr;
        scc->rx_ptr     = (ptr->l1frame).frame;
        scc->rx_len     = 0;
    }

    /* Here is the buffer where to put the "payload" */
    if (scc->rx_len < MAXFLEN)
    {
        *(scc->rx_ptr) = rr8;
        scc->rx_len++;
        scc->rx_ptr++;
    }
    else
    {
        /* Ai, the buffer is not that big! We have to skip this frame :-( */
        rxdiscard(scc); /* thow away */
        scc->rx_err = TRUE;
    }
}
/* ------------------------------------------------------------------------- */
/* int_sp, handle special condition interrupt of the SCC card */
static void int_sp (scc_t *scc)
{
    byte rr1;

    rr1 = RR (1);
    RR8;

    if (rr1 & RX_OVR)
    {
        statistics[scc->chan].rx_overrun++;
        scc->rx_err = TRUE;
    }

    if (rr1 & SDLC_EOF)
    {
        if ( (scc->rx_err == FALSE)
             &&
             (scc->rx_len >= 9) /* should have some contents... */
             &&
             ((rr1 & (CRC_ERR | RESIDUE)) == RES8)
           )
        {
            rxeof (scc);
        }
        else
        {
            rxdiscard (scc);
        }
        scc->rx_err = FALSE;
    }

    WR0 (RES_ERR);
}
/* ------------------------------------------------------------------------- */
/* timer - handle timers */
static void timer(u16 ticks)
{
    scc_t *scc;

    /* Handle running timers, 10ms resolution */
    for (scc = scctab; scc < sccend; scc++)
    {
        if ((scc->mode & MODE_off) == 0)
        {
            idisable();
            if (scc->tx_timer != (i16) 0x8000) /* running? */
            {
                if (scc->tx_timer <= ticks)
                {
                    scc->tx_timer = 0x8000;

                    switch (scc->tx_state)
                    {
                    case DELAY:
                        /* end of TX delay, get started with data now! */
                        WR0 (RES_TX_CRC);
                        WR8 (txget (scc));
                        WR0 (RES_EOM);
                        scc->tx_state = ACTIVE;
                        break;
                    case CALIBRATE:
                        scc->cal_timer--;
                        if(scc->cal_timer != 0)
                        {
                            /* another minute to go */
                            scc->tx_timer = 6000;
                            break;
                        }
                        /* fall thru */
                    case TAIL:
                        /* finished, switch off the transmitter */
                        scc_tx (scc, FALSE);
                        scc->tx_state = IDLE;
                    }
                }
                else
                {
                    scc->tx_timer -= ticks;
                }
            }
            irestore();
        }
    }
}
/* ------------------------------------------------------------------------- */
/* int_1C timer interrupt for txdelay */
static void interrupt far int_1C (void)
{
    static u16 ticks;

    if(ticks == 5)
        ticks = 6;
    else
        ticks = 5;

    if(driver_initialised == TRUE)
        timer (ticks);

    (*old_int_1C)();
}
/* ------------------------------------------------------------------------- */
/* scc_detect - Check if the SCC card is realy there */
static bool scc_detect (void)
{
    scc_t *scc;

    for (scc = scctab; scc < sccend; scc++)
    {
        RR0;
        if (   (WR (13, 0x55), RR (13) != 0x55)
            || (WR (13, 0xAA), RR (13) != 0xAA))
            return FALSE;
    }

    return TRUE;
}
/* ------------------------------------------------------------------------- */
/* init_device - Process the command line */
u16 init_device(i16 argc, char *argv[])
{
    byte       c, channel;
    byte       *cp;
    bool       par_error;
    u16        sccbase;
    u16        multiplier;
    i16        cnt;
    sccboard_t *board;
    scc_t      *scc;

    /* Copyright notice */
    pstr("\nFlexNet XSCC driver V");
    pstr(version);
    pstr(" (C) ");
    pstr(__DATE__);
    pstr(" H. de Groot PE1DNN\n");

    /* Process the command line; the order of the parameters is arbitrary! */
    par_error = FALSE;
    sccbase   = 0;
    tail_time = 0;
    for (cnt = 1; cnt < argc; ++cnt)
    {
        cp = argv[cnt];
        if (*cp == '-' || *cp == '/')
            cp++;
        else
        {
                pstr("\nCommand line error, "
                     "options shall start with '-' or '/'");
                par_error = 1;
                break;
        }

        switch (c = tolower(*cp++))
        {
            default:
                pstr("\nUnknown option '"); pch(c); pch('\'');
                par_error = 1;
                break;
            case 't':
                if((*cp == ':') || (*cp == '=')) cp++;
                switch (tolower(*cp)) {
                case 'o':   scctype = OSCC;
                            break;
                case 'p':   scctype = PSCC;
                            break;
                case 'f':   scctype = FSCC;
                            break;
                case 'd':   if(tolower(cp[1]) == 'r')
                                scctype = DRSI;
                            else
                                scctype = DSCC;
                            break;
                case 'u':   scctype = USCC;
                            break;
                default:    pstr("\nIllegal type, expected OSCC, PSCC, FSCC,"
                                 " DRSI, DSCC or USCC");
                            par_error = 1;
                            break;
                }
                break;
            case 'p':
                if((*cp == ':') || (*cp == '=')) cp++;
                sccbase = (u16) strtoul(cp, NULL, 0);
                if(sccbase < 0x100 || sccbase > 0x3F8 || sccbase & 7)
                {
                    pstr("\nIllegal addess, expect value between 0x100 and "
                         "0x3F8 on an 8 byte boundary");
                    par_error = 1;
                }
                break;
            case 'i':
                if((*cp == ':') || (*cp == '=')) cp++;
                sccirq = atoi(cp);
                if(sccirq  < 2 || sccirq == 6 || sccirq == 8
                   || sccirq == 13 || sccirq  > 15)
                {
                    pstr("\nIllegal irq, choose one of 2, 3, 4, 5, 7, 9, "
                         "10, 11, 12, 14 or 15");
                    par_error = 1;
                }
                break;
            case 'c':
                if((*cp == ':') || (*cp == '=')) cp++;
                sccchans = atoi(cp);
                break;
            case 'd':
                if((*cp == ':') || (*cp == '=')) cp++;
                tail_time = atoi(cp) * 100;
                if(tail_time > 6000)
                {
                    pstr("\nIllegal tail-delay time, expect value between "
                         "0 and 60 seconds");
                    par_error = 1;
                }
                break;
            case 'x':
                if((*cp == ':') || (*cp == '=')) cp++;
                pclk = strtoul(cp, NULL, 10);
                pclk = pclk * 10000;
                while(*cp != '\0')
                {
                    if(*cp == '.')
                    {
                        cp++;
                        break;
                    }
                    cp++;
                }

                switch(strspn(cp, "0123456789"))
                {
                    case 0:  multiplier = 0;
                             break;
                    case 1:  multiplier = 1000;
                             break;
                    case 2:  multiplier = 100;
                             break;
                    case 3:  multiplier = 10;
                             break;
                    case 4:  multiplier = 1;
                             break;
                    default: multiplier = 0;
                             break;
                }

                if(multiplier != 0)
                {
                    pclk = pclk + (strtoul(cp, NULL, 10) * multiplier);
                }

                if(pclk < 10000)
                {
                    pstr("\nIllegal Xtal, must be at least 1.0 Mhz");
                    par_error = 1;
                }
                break;
            case '?':
            case 'h':
                pstr(helptext);
                return 99;
            }
    }

    if(par_error == 0)
    {
        /* check the parameters and eventually return an error code! */
        board = &sccboard[scctype];

        max_sccchans = board->chans;

        if(sccchans == (MAX_SCCCHANS + 1))
            sccchans = board->chans;

        if((sccchans == 0) || (sccchans > max_sccchans))
        {
            pstr("\nIllegal number of channels specified");
            par_error = 1;
        }
        sccchans--; /* sccchans is base 0 */
    }

    if(par_error == 1)
    {
        pstr(helptext);
        return 1;
    }

    if(sccbase == 0)
        sccbase = board->base;

    if(sccirq == 0)
        sccirq = board->irq;

    for (channel = 0, scc = scctab; channel < max_sccchans; channel++)
    {
        scc->chan   = channel;
        scc->data   = sccbase + board->data[channel];
        scc->ctrl   = scc->data + board->ctrl;

        /* avoid 0, set to 1200 as default! */
        scc->baud = 12;

        if(channel <= sccchans)
        {
            if((scctype == OSCC) || (scctype == FSCC))
            {
                scc->mode   = MODE_c | MODE_r;
                scc->modem  = PA0HZP;
            }
            else
            {
                scc->mode   = MODE_c;
                scc->modem  = AFSK;
            }
            scc->active = TRUE;
        }
        else
        {
            scc->mode   = MODE_off;
            scc->modem  = DISABLE;
            scc->active = FALSE; /* cannot be switched on */
        }

        scc->rx_pending      = NULL;
        scc->rx_pending_last = NULL;
        scc->rx_receive      = NULL;
        scc->rx_to_flexnet   = NULL;
        scc->rx_len          = 0;
        scc->rx_ptr          = NULL;

        scc->tx_pending      = NULL;
        scc->tx_pending_last = NULL;
        scc->tx_transmit     = NULL;
        scc->tx_len          = 0;
        scc->tx_ptr          = NULL;
        scc++;
    }

    sccend = &scctab[(sccchans + 2) & 0xFE];

    /* dump what we have collected */
    if(scc_detect() == TRUE)
        show_config();
    else
    {
        pstr("\nError: ");
        pstr(card_type());
        pstr("not found on given addres!\n");
        return 1;
    }

    /* initialise the buffers */
    init_buffers();

    /* Everything ok */
    return 0;
}
/* ------------------------------------------------------------------------- */
byte l1_ch_active(byte channel)
{
    return ((scctab[channel].mode & MODE_off) == 0);
}
/* ------------------------------------------------------------------------- */
void l1_exit(void)
{
    byte       channel;
    scc_t      *scc;

    scc_hw_reset ();

    if(interrupts_initialised == TRUE)
    {
        irq_reset (sccirq, old_sccint);
        setvect (0x1C, old_int_1C);

        interrupts_initialised = FALSE;
    }
}
/* ------------------------------------------------------------------------- */
u16 l1_get_ch_cnt(void)
{
    return sccchans + 1;
}
/* ------------------------------------------------------------------------- */
static void reset_driver(scc_t *scc)
{
    i16     index;
    pool_t *ptr;

    /*
     *  reset driver variables - will flush the buffers, stop timers
     *  and reset the TX state-machine.
     */
    scc->tx_timer        = 0; /* switch off running timers */
    scc->cal_timer       = 0;
    scc->tx_state        = IDLE;  /* TX state to IDLE */
    scc->rx_err          = FALSE; /* No error (buffers are flushed) */

    while(scc->rx_pending != NULL)
    {
        ptr = scc->rx_pending;
        scc->rx_pending = ptr->next;
        rx_free(ptr);
    }
    scc->rx_pending_last = NULL;
    if(scc->rx_to_flexnet != NULL)
    {
        rx_free(scc->rx_to_flexnet);
        scc->rx_to_flexnet = NULL;
    }
    if(scc->rx_receive != NULL)
    {
        rx_free(scc->rx_receive);
        scc->rx_receive = NULL;
    }
    scc->rx_len          = 0;
    scc->rx_ptr          = NULL;

    while(scc->tx_pending != NULL)
    {
        ptr = scc->tx_pending;
        scc->tx_pending = ptr->next;
        tx_free(ptr);
    }
    scc->tx_pending_last = NULL;
    if(scc->tx_transmit != NULL)
    {
        tx_free(scc->tx_transmit);
        scc->tx_transmit = NULL;
    }
    scc->tx_len          = 0;
    scc->tx_ptr          = NULL;
}
/* ------------------------------------------------------------------------- */
static void l1_init_all(void)
{
    scc_t *scc;
    i16    index;

    scc_hw_reset ();

    if(interrupts_initialised == FALSE)
    {
        old_sccint = irq_init (sccirq, scc_int);

        /* get timer-interrupt for TX-Delay */
        old_int_1C  = getvect (0x1C);
        setvect (0x1C, int_1C);

        interrupts_initialised = TRUE;
    }

    for (scc = scctab; scc < sccend; scc++)
    {
        if ((scc->mode & MODE_off) == 0)
        {
            /* reset driver variables */
            reset_driver(scc);

            WR (4, SDLC);
            WR (10, (scc->mode & MODE_z) ? NRZ | ABUNDER : NRZI | ABUNDER);
            WR (7, SDLC_FLAG);
            WR (3, RX_DIS);
            WR (5, TX_DIS);

            scc_clk (scc, FALSE);

            WR (3, RX_EN);
            WR (5, scc->modem == AFSK ? TX_DIS : TX_EN);

            WR (15, SYNC_IE | DCD_IE);
            WR0 (RES_EX_INT);
            WR0 (RES_EX_INT);
            WR (1, EX_IE | TX_IE | RX_IE);
        }

        scc->rr0 = RR0;
    }

    for (scc = scctab; scc < sccend; scc += 2)
    {
        WR (2, scc->chan / 2 * 0x10);
        WR (9, MINT_EN);
    }

    /* whole card is only done once, the different channels  */
    /* may be changed later however...                       */
    /* during imergency this variable may become FALSE again */
    driver_initialised = TRUE;
}
/* ------------------------------------------------------------------------- */
byte l1_init_kanal(byte channel, u16 baud, u16 mode)
{
    scc_t *scc;

    idisable();

    /* first change settings for this channel, then restart the channel */
    /* currently only support for 'on' and 'off' */
    scc = &scctab[channel];

    /* reset some driver variables for this channel */
    reset_driver(scc);

    /* store mode parameter if port is usable */
    if(scc->active != TRUE) /* only if port is active */
    {
        irestore();
        return 0;           /* finished because port is not active! */
    }

    /* set modem type */
    switch(mode & (MODE_r | MODE_t))
    {
    case 0:                 scc->modem = AFSK;   /* RX and TX internal     */
                            break;
    case MODE_r:            scc->modem = PA0HZP; /* RX internal            */
                            break;               /* TX clock will be RX/32 */
    case MODE_t:            scc->modem = G3RUH;  /* RX internal            */
                            break;               /* TX external            */
    case MODE_r | MODE_t:   scc->modem = DF9IC;  /* RX and TX external     */
                            break;               /* also needs MODE_z      */
    }

    if(mode & MODE_off)
        scc->modem = DISABLE;

    /* if non-zero change, otherwise keep old value */
    if(baud != 0)
        scc->baud = baud;

    if(driver_initialised == FALSE)
    {
        l1_init_all(); /* intialise the whole card */
    }

    /*
     * Switch port off to silence this port so we are free to
     * reinitalise it if needed.
     */
    /* disable interrupts */
    WR (15, 0);
    WR0 (RES_EX_INT);
    WR0 (RES_EX_INT);
    WR (1, 0);

    /* disable receiver and transmitter */
    WR (3, RX_DIS);
    WR (5, TX_DIS);

    /* switch off PTT - will also reprogram the clock (baud-rate) */
    scc_clk (scc, FALSE);

    if ((mode & MODE_off) == 0)
    {
        if((scc->mode & MODE_z) != (mode & MODE_z))
        {
            /* change from NRZ or NRZI or vice-versa */
            WR (10, (mode & MODE_z) ? NRZ | ABUNDER : NRZI | ABUNDER);
        }

        /* re-enable the interrupts */
        WR (3, RX_EN);
        WR (5, scc->modem == AFSK ? TX_DIS : TX_EN);

        /*
         * this channel was switched off but is switched on now,
         * enable interrupt generation on various events
         */
        WR (15, SYNC_IE | DCD_IE);
        WR0 (RES_EX_INT);
        WR0 (RES_EX_INT);
        WR (1, EX_IE | TX_IE | RX_IE);

        scc->mode = mode;
        irestore();
        return 1;
    }
    else
    {
        /*
         * shall not re-enable interrupts
         * this channel shall stay off
         */
        scc->mode = mode;
        irestore();
        return 0;
    }
}
/* ------------------------------------------------------------------------- */
#pragma argsused
char far *l1_ident(byte channel)
{
    static char far *id = "x:XSCC0";

    switch(scctype) {
    case OSCC: id[0] = 'O'; break;
    case PSCC: id[0] = 'P'; break;
    case FSCC: id[0] = 'F'; break;
    case DRSI: id[0] = 'D'; id[1] = 'R'; id[2] = 'S'; id[3] = 'I'; break;
    case DSCC: id[0] = 'D'; break;
    case USCC: id[0] = 'U'; break;
    }

    id[6] = '1' + sccchans;
    return id;
}
/* ------------------------------------------------------------------------- */
#pragma argsused
char far *l1_version(byte kanal)
{
    return version;
}
/* ------------------------------------------------------------------------- */
L1_STATISTICS far *l1_stat(byte channel, byte delete)
{
    L1_STATISTICS *p = &statistics[channel];
    if (delete)
        p->tx_error = p->rx_overrun = p->rx_bufferoverflow =
        p->tx_frames = p->rx_frames = p->io_error = 0;
    return p;
}
/* ------------------------------------------------------------------------- */
byte l1_ch_state(byte channel)
{
    scc_t *scc;
    byte   status;

    scc = &scctab[channel];

    /* don't allow changes while we are busy */
    idisable();

    /* Start with cleared status bits */
    status = 0;

    /* RX data pending? */
    if(scc->rx_pending != NULL)
    {
        /*
         * DCD shall overlap RXB, so report DCD even if the 'real' DCD
         * condition does not exist
         */
        status = status | (CH_RXB | CH_DCD);

        /*
         * Remember this channel, check this on l1_rx_frame!
         */
        rx_channel = channel;
    }

    /* PTT on? */
    if(scc->tx_state != IDLE)
        status = status | CH_PTT;

    /* DCD active ? */
    if(scc->dcd)
        status = status | CH_DCD;

    /* CALIBRATING? */
    if(scc->tx_state == CALIBRATE)
        status = status | CH_TBY;

    /* Full Duplex? */
    if(scc->mode & MODE_d)
        status = status | CH_FDX;

    /* Return to previous interrupt setting */
    irestore();

    return status;
}
/* ------------------------------------------------------------------------- */
u16 l1_scale(byte channel)
{
    scc_t *scc;

    scc = &scctab[channel];

    return 6144 / scc->baud;
}
/* ------------------------------------------------------------------------- */
void l1_tx_calib(byte channel, byte minutes)
{
    scc_t   *scc;

    if(driver_initialised == FALSE) return;

    scc = &scctab[channel];

    if((scc->mode & MODE_off) == 0)
    {
        if(minutes != 0)
        {
            idisable ();
            if( scc->tx_state == IDLE)
            {
                /*
                 * State is IDLE, we need to key-up the transmitter
                 */
                scc_tx (scc, TRUE); /* PTT on, airborne... */
            }
            scc->cal_timer = minutes;
            scc->tx_timer  = 6000; /* at least one minute */
            scc->tx_state  = CALIBRATE;
            irestore ();
        }
        else
        {
            /*
             * if minutes is 0 then terminate CALIBRATE
             * if it was in progress.
             */
            idisable ();
            if( scc->tx_state == CALIBRATE)
            {
                /*
                 * State is CALIBRATE, we need to stop calibration
                 */
                scc_tx (scc, FALSE); /* PTT off, off-air... */
                scc->cal_timer = 0; /* switch timers off */
                scc->tx_timer  = 0x8000;
                scc->tx_state  = IDLE;
            }
            irestore ();
        }
    }
}
/* ------------------------------------------------------------------------- */
#pragma argsused
void set_led(byte kanal, byte ledcode)
{
}
/* ------------------------------------------------------------------------- */
/* receive an AX25 frame */
L1FRAME far *l1_rx_frame()
{
    scc_t  *scc;
    pool_t *ptr;
    byte   channel;

    if(driver_initialised == FALSE)
        return (L1FRAME far *) NULL;

    /* don't allow changes while we are busy */
    idisable();

    /*
     * First check the SCC channel for which l1_ch_state is called
     * the last time. Then cycle through the other channels if no
     * pending data is found. As soon as pending data is found the
     * loop is terminated.
     */
    scc = &scctab[rx_channel];

    channel = 0xff; /* first increment shall give 0 */
    do {
        if((scc->mode & MODE_off) == 0)
        {
            /* reclaim buffer given last time */
            if(scc->rx_to_flexnet != NULL)
            {
                rx_free(scc->rx_to_flexnet);
                scc->rx_to_flexnet = NULL;
            }

            if((ptr = scc->rx_pending) != NULL)
            {
                /* this buffer contains a payload to send back   */
                /* swap pointer and return ptr to this rx_buffer */
                scc->rx_pending    = ptr->next;
                scc->rx_to_flexnet = ptr;
                /* return immediately with the found data */
                irestore();
                return &(ptr->l1frame);
            }
        }
        channel++;
        scc = &scctab[channel];
    } while (channel <= sccchans);

    /* no data pending... */
    irestore();
    return (L1FRAME far *) NULL;
}
/* ------------------------------------------------------------------------- */
/* deliver not-used TX buffer */
L1FRAME far *l1_get_framebuf(byte channel)
{
    if(driver_initialised == FALSE)
        return (L1FRAME far *) NULL;

    /* don't allow changes while we are busy */
    idisable();

    tx_channel = channel; /* remember channel for last fetched TX buffer */

    if(tx_to_flexnet == NULL)
    {
        /* Get a new buffer to give to FlexNet */
        if(next_free_tx == NULL)
        {
            /* out of transmit buffers */
            irestore();
            return (L1FRAME far*) NULL;
        }
        /* Allocate transmit buffer */
        tx_to_flexnet = next_free_tx;
        next_free_tx  = next_free_tx->next;
    }

    /* Return the same or new buffer as last time to FlexNet */
    irestore();
    return &(tx_to_flexnet->l1frame);
}
/* ------------------------------------------------------------------------- */
/* transmit the frame which was fetched last */
byte l1_tx_frame(void)
{
    byte    txdelay;
    scc_t  *scc;

    if(driver_initialised == FALSE)
        return 0;

    /* don't allow changes while we are busy */
    idisable();

    scc = &scctab[tx_channel];

    /* check if this can be transmitted now */
    if((scc->mode & MODE_off) == 0)
    {
        if(tx_to_flexnet == NULL)
        {
            /* call to l1_tx_frame without buffer fetch ?!? */
            irestore();
            return 0;
        }

        /*
         * add buffer to transmit list, it contains data for transmission
         */

        /* the next pointer should not point to anything */
        tx_to_flexnet->next = NULL; /* this will become the end of the chain */

        if(scc->tx_pending == NULL)
        {
            /* first buffer in tx_pending list */
            scc->tx_pending = tx_to_flexnet;
        }
        else
        {
            /* add buffer to tx_pending list */
            (scc->tx_pending_last)->next = tx_to_flexnet;
        }
        /* this is the last frame we added */
        scc->tx_pending_last = tx_to_flexnet;

        switch (scc->tx_state) {
        case IDLE:
            /*
             * State is IDLE, we need to key-up the transmitter
             */
            txdelay = (tx_to_flexnet->l1frame).txdelay;
            scc->tx_timer = txdelay == 0 ? 25 : txdelay;
            scc->tx_state = DELAY;
            scc_tx (scc, TRUE); /* PTT on, airborne... */
            break;
        case TAIL:
            scc->tx_timer = 1;
            scc->tx_state = DELAY;
            break;
        }

        /* next time we have to allocate a new one */
        tx_to_flexnet = NULL;

        /* transmitting... */
        irestore();
        return 1;
    }
    else
    {
        /* we are done, function can be called again */
        irestore();
        return 0;
    }
}
/*------------------------------------------------------------------------- */
